/*                                                              -*-c++-*-
 * Copyright (c) 2015-2019,2021-2023 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Anthony Mallet on Wed Sep 16 2015
 */
#ifndef H_POM_CODELS
#define H_POM_CODELS

#include <aio.h>

#include <set>

#include "pom_c_types.h"
#include "ukf.h"

struct pom_snapshot {
  EIGEN_MAKE_ALIGNED_OPERATOR_NEW;

  mutable bool fused;

  mutable ukf::state_s state;
  mutable ukf::state_s::dcov_s cov;

  pom_measurement_s measurement;

  bool
  operator<(const pom_snapshot &other) const
  {
    /* sorted by greater timestamps first */
    if (measurement.data.ts.sec > other.measurement.data.ts.sec)
      return true;
    if (measurement.data.ts.sec == other.measurement.data.ts.sec)
      if (measurement.data.ts.nsec > other.measurement.data.ts.nsec)
        return true;
    return false;
  }
};

typedef std::multiset<
  pom_snapshot,
  std::less<pom_snapshot>,
  Eigen::aligned_allocator<pom_snapshot>
  > pom_history_s;

struct pom_context_s {
  EIGEN_MAKE_ALIGNED_OPERATOR_NEW;

  ukf::state_s zero;
  ukf::state_s::dcov_s zerocov;

  or_time_ts ts;
  ukf::state_s state;
  ukf::state_s::dcov_s cov;

  ukf::filter_s filter;
  pom_history_s history;
};

struct pom_log_s {
  int fd;
  struct aiocb req;
  char buffer[4096];
  bool pending, skipped;
  uint32_t decimation;
  size_t missed, total;

# define pom_logfmt	" %g"

# define pom_log_header_fmt                                             \
  "ts i posp attp velp avelp accp "                                     \
  "x y z roll pitch yaw vx vy vz wx wy wz ax ay az "                    \
  "sx2 sy2 sz2 sr2 sp2 sh2 svx2 svy2 svz2 swx2 swy2 swz2 sax2 say2 saz2"
# define pom_log_line_fmt                                               \
  "%d.%09d %d %d %d %d %d %d"                                           \
  /* p */ pom_logfmt pom_logfmt pom_logfmt                              \
  /* rpy */ pom_logfmt pom_logfmt pom_logfmt                            \
  /* v */ pom_logfmt pom_logfmt pom_logfmt                              \
  /* w */ pom_logfmt pom_logfmt pom_logfmt                              \
  /* a */ pom_logfmt pom_logfmt pom_logfmt                              \
  /* sp */ pom_logfmt pom_logfmt pom_logfmt                             \
  /* sq */ pom_logfmt pom_logfmt pom_logfmt                             \
  /* sv */ pom_logfmt pom_logfmt pom_logfmt                             \
  /* sw */ pom_logfmt pom_logfmt pom_logfmt                             \
  /* sa */ pom_logfmt pom_logfmt pom_logfmt

# define pom_log_m_header_fmt                                           \
  "mid delay ts i posp attp velp avelp accp "                           \
  "x y z qw qx qy qz vx vy vz wx wy wz ax ay az "                       \
  "offx offy offz offr offp offh"
# define pom_log_m_line_fmt                                             \
  "%d" pom_logfmt " %d.%09d %d %d %d %d %d %d"                          \
  /* p */ pom_logfmt pom_logfmt pom_logfmt                              \
  /* q */ pom_logfmt pom_logfmt pom_logfmt pom_logfmt                   \
  /* v */ pom_logfmt pom_logfmt pom_logfmt                              \
  /* w */ pom_logfmt pom_logfmt pom_logfmt                              \
  /* a */ pom_logfmt pom_logfmt pom_logfmt                              \
  /* offp */ pom_logfmt pom_logfmt pom_logfmt                           \
  /* offq */ pom_logfmt pom_logfmt pom_logfmt
# define pom_scan_m_line_fmt                                            \
  "%u %*f %d.%09d %d %d %d %d %d %d"                                    \
  /* p, q */ "%lf %lf %lf %lf %lf %lf %lf"                              \
  /* v, w */ "%lf %lf %lf %lf %lf %lf"                                  \
  /* a */ "%lf %lf %lf"                                                 \
  /* off.p, off.q */ "%lf %lf %lf %lf %lf %lf"
};


static inline genom_event
pom_e_sys_error(const char *s, genom_context self)
{
  pom_e_sys_detail d;
  char buf[64], *p;

  d.code = errno;

#ifdef STRERROR_R_CHAR_P
  /* glibc managed to mess up with this function */
  p = strerror_r(d.code, buf, sizeof(buf));
#else
  strerror_r(d.code, buf, sizeof(buf));
  p = buf;
#endif
  snprintf(d.what, sizeof(d.what), "%s%s%s", s ? s : "", s ? ": " : "", p);

  return pom_e_sys(&d, self);
}

#endif /* H_POM_CODELS */
