/*
 * Copyright (c) 2015-2023 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Anthony Mallet on Mon Aug 24 2015
 */
#include "acpom.h"

#include <fcntl.h>
#include <unistd.h>

#include <cstdio>

#include <Eigen/Geometry>

#include "pom_c_types.h"

#include "codels.h"
#include "ukf.h"


/* --- Attribute set_filter_period -------------------------------------- */

/** Validation codel set_filter_period of attribute set_filter_period.
 *
 * Returns genom_ok.
 * Throws pom_e_sys.
 */
genom_event
set_filter_period(uint32_t fperiodms, const genom_context self)
{
  if (fperiodms < pom_period_ms) {
    errno = EINVAL;
    return pom_e_sys_error("set_filter_period", self);
  }

  return genom_ok;
}


/* --- Attribute set_prediction_model ----------------------------------- */

/** Validation codel set_prediction_model of attribute set_prediction_model.
 *
 * Returns genom_ok.
 * Throws .
 */
genom_event
set_prediction_model(pom_prediction_model pmodel,
                     pom_log_s **log_state, const genom_context self)
{
  const char *strmodel = "unknown";

  switch (pmodel) {
    case pom_constant_velocity:
      ukf::state_s::pmodel = ukf::state_s::constant_velocity;
      strmodel = "constant_velocity";
      break;

    case pom_constant_acceleration:
      ukf::state_s::pmodel = ukf::state_s::constant_acceleration;
      strmodel = "constant_acceleration";
      break;
  }

  /* log it */
  if ((*log_state)->fd >= 0)
    dprintf((*log_state)->fd, "# process model %s\n", strmodel);

  return genom_ok;
}


/* --- Attribute set_process_noise -------------------------------------- */

/** Validation codel set_process_noise of attribute set_process_noise.
 *
 * Returns genom_ok.
 * Throws .
 */
genom_event
set_process_noise(double max_jerk, double max_dw,
                  pom_log_s **log_state, const genom_context self)
{
  ukf::state_s::max_dadt = max_jerk;
  ukf::state_s::max_dwdt = max_dw;

  /* log it */
  if ((*log_state)->fd >= 0)
    dprintf(
      (*log_state)->fd, "# process noise acc %g m/s³, aacc %g rad/s²\n",
      max_jerk, max_dw);

  return genom_ok;
}


/* --- Attribute set_mag_field ------------------------------------------ */

/** Validation codel set_mag_field of attribute set_mag_field.
 *
 * Returns genom_ok.
 * Throws .
 */
genom_event
set_mag_field(const pom_ids_mag_field *magdir, pom_log_s **log_state,
              const genom_context self)
{
  ukf::measure_s::magdir =
    Eigen::Matrix<double, 3, 1>(magdir->x, magdir->y, magdir->z).normalized();

  /* log it */
  if ((*log_state)->fd >= 0)
    dprintf(
      (*log_state)->fd, "# magnetic field X %g µT, Y %g µT, Z %g µT\n",
      magdir->x * 1e6, magdir->y * 1e6, magdir->z * 1e6);

  return genom_ok;
}


/* --- Activity log_state ----------------------------------------------- */

/** Validation codel pom_log_open of activity log_state.
 *
 * Returns genom_ok.
 * Throws pom_e_sys.
 */
genom_event
pom_log_open(const char path[64], pom_log_s **log,
             const genom_context self)
{
  int fd;

  fd = open(path, O_WRONLY|O_APPEND|O_CREAT|O_TRUNC, 0666);
  if (fd < 0) return pom_e_sys_error(path, self);

  if ((*log)->fd >= 0)
    close((*log)->fd);
  if ((*log)->req.aio_fildes >= 0) {
    if ((*log)->pending)
      while (aio_error(&(*log)->req) == EINPROGRESS)
        /* empty body */;
  }
  (*log)->fd = fd;
  (*log)->req.aio_fildes = -1;
  (*log)->pending = false;
  (*log)->skipped = false;
  (*log)->decimation = 1;
  (*log)->missed = 0;
  (*log)->total = 0;

  return genom_ok;
}


/* --- Activity log_measurements ---------------------------------------- */

/** Validation codel pom_log_open of activity log_measurements.
 *
 * Returns genom_ok.
 * Throws pom_e_sys.
 */
/* already defined in service log_state validation */



/* --- Function set_position -------------------------------------------- */

/** Codel pom_set_position of function set_position.
 *
 * Returns genom_ok.
 */
genom_event
pom_set_position(double x, double y, double z, double roll,
                 double pitch, double yaw, double precision,
                 pom_context_s **context, const genom_context self)
{
  double cr = cos(roll * M_PI/180./2.),  sr = sin(roll * M_PI/180./2.);
  double cp = cos(pitch * M_PI/180./2.), sp = sin(pitch * M_PI/180./2.);
  double cy = cos(yaw * M_PI/180./2.),   sy = sin(yaw * M_PI/180./2.);

  (*context)->zero.p() << x, y, z;
  (*context)->zero.q() <<
    sr * cp * cy - cr * sp * sy, /* x */
    cr * sp * cy + sr * cp * sy, /* y */
    cr * cp * sy - sr * sp * cy, /* z */
    cr * cp * cy + sr * sp * sy; /* w */

  (*context)->zero.v() << 0., 0., 0.;
  (*context)->zero.w() << 0., 0., 0.;

  (*context)->zero.a() << 0., 0., 0.;

  (*context)->zerocov =
    ukf::state_s::dcov_s::Identity() * precision*precision/9;

  /* reset measurements */
  (*context)->history.clear();

  return genom_ok;
}


/* --- Function log_stop ------------------------------------------------ */

/** Codel pom_log_stop of function log_stop.
 *
 * Returns genom_ok.
 */
genom_event
pom_log_stop(pom_log_s **log_state, pom_log_s **log_measurements,
             const genom_context self)
{
  if (*log_state && (*log_state)->req.aio_fildes >= 0)
    close((*log_state)->req.aio_fildes);
  (*log_state)->fd = (*log_state)->req.aio_fildes = -1;

  if (*log_measurements && (*log_measurements)->req.aio_fildes >= 0)
    close((*log_measurements)->req.aio_fildes);
  (*log_measurements)->fd = (*log_measurements)->req.aio_fildes = -1;

  return genom_ok;
}


/* --- Function log_info ------------------------------------------------ */

/** Codel pom_log_info of function log_info.
 *
 * Returns genom_ok.
 */
genom_event
pom_log_info(const pom_log_s *log_state,
             const pom_log_s *log_measurements, double *state_miss,
             double *measurements_miss, const genom_context self)
{
  *state_miss =
    log_state->total ?
    100.0 * log_state->missed / log_state->total : 0.;
  *measurements_miss =
    log_measurements->total ?
    100.0 * (double)log_measurements->missed / log_measurements->total : 0.;

  return genom_ok;
}
