/*
 * Copyright (c) 2015-2023 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Anthony Mallet on Mon Aug 24 2015
 */
#include "acpom.h"

#include <sys/time.h>
#include <aio.h>
#include <err.h>
#include <unistd.h>

#include <cstdio>
#include <cmath>

#include <Eigen/Geometry>

#include "pom_c_types.h"

#include "codels.h"
#include "ukf.h"


/* --- Task filter ------------------------------------------------------ */

/** Codel pom_filter_start of task filter.
 *
 * Triggered by pom_start.
 * Yields to pom_exec.
 * Throws pom_e_sys.
 */
genom_event
pom_filter_start(pom_context_s **context, const pom_frame *frame,
                 const genom_context self)
{
  frame->open("robot", self);
  return pom_exec;
}


/** Codel pom_filter_exec of task filter.
 *
 * Triggered by pom_exec.
 * Yields to pom_pause_exec.
 * Throws pom_e_sys.
 */
genom_event
pom_filter_exec(pom_context_s **context, uint32_t fperiodms,
                pom_log_s **log_state, const pom_frame *frame,
                const genom_context self)
{
  static struct timeval last;
  static double avgperiod;
  static uint32_t slow;

  if (!(*context)) return pom_pause_exec;

  const pom_history_s &h = (*context)->history;
  pom_history_s::reverse_iterator todo, skip, i, j;
  struct timeval start, now;
  double date, mdate;

  /* measure start time and wait for next period if needed */
  gettimeofday(&start, NULL);
  if ((start.tv_sec - last.tv_sec) * 1e3 +
      (start.tv_usec - last.tv_usec) * 1e-3 < fperiodms - pom_period_ms/2.)
    return pom_pause_exec;
  last = start;

  /* get the oldest non fused state */
  for(todo = h.rbegin(); todo != h.rend(); ++todo)
    if (!todo->fused) break;

  /* set filter state to the fused state before - or initial state when no
   * measurements are present */
  if (todo == h.rbegin()) {
    (*context)->filter.set((*context)->zero, (*context)->zerocov);
    date = (*context)->ts.sec + 1e-9 * (*context)->ts.nsec;
  } else {
    todo--;
    (*context)->filter.set(todo->state, todo->cov);
    date = todo->measurement.data.ts.sec +
           1e-9 * todo->measurement.data.ts.nsec;
    todo++;
  }

  /* process history */
  skip = h.rend();
  for(; todo != h.rend(); ++todo) {
    const pom_snapshot &s = *todo;
    s.fused = false;

    /* if we've been computing for more than the period, skip to the
     * last measurement to be fused - skipped items will be processed next
     * time */
    if (skip == h.rend()) {
      gettimeofday(&now, NULL);
      if ((now.tv_sec - start.tv_sec) * 1e3 +
          (now.tv_usec - start.tv_usec) * 1e-3 > fperiodms) {
        for(i = j = todo, j++; j != h.rend(); ++i, ++j) {
          if (i->fused && !j->fused) skip = i;
        }
        if (skip == h.rend()) {
          /* nothing can be skipped: continue as normal + disable the test */
          skip = h.rbegin();
        } else {
          todo = skip;
          (*context)->filter.set(todo->state, todo->cov);
          date = todo->measurement.data.ts.sec +
                 1e-9 * todo->measurement.data.ts.nsec;
          continue;
        }
      }
    }

    mdate = s.measurement.data.ts.sec + 1e-9 * s.measurement.data.ts.nsec;

    if (mdate < date) {
      /*
       * skip measurements older than current state - this should not happen,
       * except for the very first measurement in the component lifetime.
       */
      (*context)->filter.predict(0.);
      (*context)->filter.update(s.state, s.cov);
      (*context)->filter.set(s.state, s.cov);
      s.fused = true;
      continue;
    }

    (*context)->filter.predict(mdate - date);

    /* position */
    if (s.measurement.data.pos._present && !s.measurement.data.intrinsic &&
        !std::isnan(s.measurement.data.pos._value.x)) {
      (*context)->filter.measure.p() <<
        s.measurement.data.pos._value.x,
        s.measurement.data.pos._value.y,
        s.measurement.data.pos._value.z;

      if (s.measurement.data.pos_cov._present) {
        const double *c = s.measurement.data.pos_cov._value.cov;

        (*context)->filter.measure.p_cov() <<
          c[0],  c[1],  c[3],
          c[1],  c[2],  c[4],
          c[3],  c[4],  c[5];
      } else {
        /* no covariance provided, provide a pessimistic default */
        (*context)->filter.measure.p_cov() <<
          1.,  0.,  0.,
          0.,  1.,  0.,
          0.,  0.,  1.;
      }
    }

    /* position in body frame - assume altimeter on Z */
    if (s.measurement.data.pos._present && s.measurement.data.intrinsic &&
        !std::isnan(s.measurement.data.pos._value.z)) {
      (*context)->filter.measure.ipz() << s.measurement.data.pos._value.z;

      if (s.measurement.data.pos_cov._present) {
        (*context)->filter.measure.ipz_cov() <<
          s.measurement.data.pos_cov._value.cov[5];
      } else {
        /* no covariance provided, provide a pessimistic default */
        (*context)->filter.measure.ipz_cov() << 1.;
      }
    }

    /* orientation */
    if (s.measurement.data.att._present && !s.measurement.data.intrinsic &&
        !std::isnan(s.measurement.data.att._value.qw)) {
      (*context)->filter.measure.q() <<
        s.measurement.data.att._value.qx,
        s.measurement.data.att._value.qy,
        s.measurement.data.att._value.qz,
        s.measurement.data.att._value.qw;

      if (s.measurement.data.att_cov._present) {
        const double *c = s.measurement.data.att_cov._value.cov;

        (*context)->filter.measure.q_cov() <<
          c[2], c[4], c[7], c[1],
          c[4], c[5], c[8], c[3],
          c[7], c[8], c[9], c[6],
          c[1], c[3], c[6], c[0];
      } else {
        /* no covariance provided, provide a pessimistic default */
        double qx = s.measurement.data.att._value.qx;
        double qy = s.measurement.data.att._value.qy;
        double qz = s.measurement.data.att._value.qz;
        double qw = s.measurement.data.att._value.qw;

        (*context)->filter.measure.q_cov() <<
          0.25*(1-qx*qx), -0.25*qx*qy,     -0.25*qx*qz,     -0.25*qw*qx,
          -0.25*qx*qy,     0.25*(1-qy*qy), -0.25*qy*qz,     -0.25*qw*qy,
          -0.25*qx*qz,    -0.25*qy*qz,      0.25*(1-qz*qz), -0.25*qw*qz,
          -0.25*qw*qx,    -0.25*qw*qy,     -0.25*qw*qz,      0.25*(1-qw*qw);
      }
    }

    /* magnetic field */
    if (s.measurement.data.att._present && s.measurement.data.intrinsic &&
        std::isnan(s.measurement.data.att._value.qw)) {
      (*context)->filter.measure.im() <<
        s.measurement.data.att._value.qx,
        s.measurement.data.att._value.qy,
        s.measurement.data.att._value.qz;

      if (s.measurement.data.att_cov._present) {
        const double *c = s.measurement.data.att_cov._value.cov;

        (*context)->filter.measure.im_cov() <<
          c[2], c[4], c[7],
          c[4], c[5], c[8],
          c[7], c[8], c[9];
      } else {
        /* no covariance provided, provide a pessimistic default */
        (*context)->filter.measure.im_cov() <<
          1e-5, 0.,    0.,
          0.,    1e-5, 0.,
          0.,    0.,    1e-5;
      }
    }

    /* linear velocity */
    if (s.measurement.data.vel._present &&
        !std::isnan(s.measurement.data.vel._value.vx)) {
      (s.measurement.data.intrinsic ?
       (*context)->filter.measure.iv() :
       (*context)->filter.measure.v()) <<
        s.measurement.data.vel._value.vx,
        s.measurement.data.vel._value.vy,
        s.measurement.data.vel._value.vz;

      if (s.measurement.data.vel_cov._present) {
        const double *c = s.measurement.data.vel_cov._value.cov;

        (s.measurement.data.intrinsic ?
         (*context)->filter.measure.iv_cov() :
         (*context)->filter.measure.v_cov()) <<
          c[0], c[1], c[3],
          c[1], c[2], c[4],
          c[3], c[4], c[5];
      } else {
        /* no covariance provided, provide a pessimistic default */
        (s.measurement.data.intrinsic ?
         (*context)->filter.measure.iv_cov() :
         (*context)->filter.measure.v_cov()) <<
          1., 0., 0.,
          0., 1., 0.,
          0., 0., 1.;
      }
    }

    /* angular velocity */
    if (s.measurement.data.avel._present &&
        !std::isnan(s.measurement.data.avel._value.wx)) {
      (s.measurement.data.intrinsic ?
       (*context)->filter.measure.iw() :
       (*context)->filter.measure.w()) <<
        s.measurement.data.avel._value.wx,
        s.measurement.data.avel._value.wy,
        s.measurement.data.avel._value.wz;

      if (s.measurement.data.avel_cov._present) {
        const double *c = s.measurement.data.avel_cov._value.cov;

        (s.measurement.data.intrinsic ?
         (*context)->filter.measure.iw_cov() :
         (*context)->filter.measure.w_cov()) <<
          c[0], c[1], c[3],
          c[1], c[2], c[4],
          c[3], c[4], c[5];
      } else {
        /* no covariance provided, provide a pessimistic default */
        (s.measurement.data.intrinsic ?
         (*context)->filter.measure.iw_cov() :
         (*context)->filter.measure.w_cov()) <<
          0.25,  0.,    0.,
          0.,    0.25,  0.,
          0.,    0.,    0.25;
      }
    }

    /* linear acceleration - XXX when intrisic, assume accelerometer */
    if (s.measurement.data.acc._present &&
        !std::isnan(s.measurement.data.acc._value.ax)) {
      (s.measurement.data.intrinsic ?
       (*context)->filter.measure.iawg() :
       (*context)->filter.measure.a()) <<
        s.measurement.data.acc._value.ax,
        s.measurement.data.acc._value.ay,
        s.measurement.data.acc._value.az;

      if (s.measurement.data.acc_cov._present) {
        const double *c = s.measurement.data.acc_cov._value.cov;

        (s.measurement.data.intrinsic ?
         (*context)->filter.measure.iawg_cov() :
         (*context)->filter.measure.a_cov()) <<
          c[0], c[1], c[3],
          c[1], c[2], c[4],
          c[3], c[4], c[5];
      } else {
        /* no covariance provided, provide a pessimistic default */
        (s.measurement.data.intrinsic ?
         (*context)->filter.measure.iawg_cov() :
         (*context)->filter.measure.a_cov()) <<
          1., 0., 0.,
          0., 1., 0.,
          0., 0., 1.;
      }
    }

    /* sensor offset */
    (*context)->filter.measure.offset.p <<
      s.measurement.offset.p.x,
      s.measurement.offset.p.y,
      s.measurement.offset.p.z;
    (*context)->filter.measure.offset.q.w() =
      s.measurement.offset.q.qw;
    (*context)->filter.measure.offset.q.vec() <<
      s.measurement.offset.q.qx,
      s.measurement.offset.q.qy,
      s.measurement.offset.q.qz;

    /* update */
    (*context)->filter.update(s.state, s.cov);
    s.fused = true;

    /* update state for next iteration */
    (*context)->filter.set(s.state, s.cov);
    date = mdate;
  }

  /* current state */
  gettimeofday(&now, NULL);
  (*context)->filter.predict(now.tv_sec + 1e-6*now.tv_usec - date);
  (*context)->filter.update((*context)->state, (*context)->cov);
  (*context)->ts.sec = now.tv_sec;
  (*context)->ts.nsec = now.tv_usec * 1000;

  avgperiod =
    (1. - 0.01) * avgperiod +
    0.01 * (
      (now.tv_sec - start.tv_sec)*1e3 + (now.tv_usec - start.tv_usec)*1e-3)
    / fperiodms;
  if (avgperiod > 1.25) {
    slow++;
    if ((slow & ((slow-1) | 254)) == 0 || slow % (1<<15) == 0) {
      warnx("slow computation for %d periods (%fms) ...",
            slow, avgperiod*fperiodms);
    }
  } else if (slow && avgperiod < 1.05) {
    warnx("recovered good computation speed after %d periods", slow);
    slow = 0;
  }

  /* publish */
  or_pose_estimator_state *r = frame->data("robot", self);
  if (r) {
    r->ts = (*context)->ts;
    r->intrinsic = false;

    r->pos._present = true;
    r->pos._value.x = (*context)->state.p()(0);
    r->pos._value.y = (*context)->state.p()(1);
    r->pos._value.z = (*context)->state.p()(2);
    r->att._present = true;
    r->att._value.qw = (*context)->state.q()(3);
    r->att._value.qx = (*context)->state.q()(0);
    r->att._value.qy = (*context)->state.q()(1);
    r->att._value.qz = (*context)->state.q()(2);
    {
      Eigen::Matrix<double, 7, 7> c =
        (*context)->state.pq_cov((*context)->cov);

      double *pc = r->pos_cov._value.cov;
      r->pos_cov._present = true;
      pc[0] = c(0, 0);
      pc[1] = c(1, 0);  pc[2] = c(1, 1);
      pc[3] = c(2, 0);  pc[4] = c(2, 1);  pc[5] = c(2, 2);

      double *qc = r->att_cov._value.cov;
      r->att_cov._present = true;
      qc[0] = c(6, 6);
      qc[1] = c(3, 6);  qc[2] = c(3, 3);
      qc[3] = c(4, 6);  qc[4] = c(4, 3);  qc[5] = c(4, 4);
      qc[6] = c(5, 6);  qc[7] = c(5, 3);  qc[8] = c(5, 4);  qc[9] = c(5, 5);

      double *pqc = r->att_pos_cov._value.cov;
      r->att_pos_cov._present = true;
      pqc[0] = c(6, 0);  pqc[1] = c(6, 1);  pqc[2] = c(6, 2);
      pqc[3] = c(3, 0);  pqc[4] = c(3, 1);  pqc[5] = c(3, 2);
      pqc[6] = c(4, 0);  pqc[7] = c(4, 1);  pqc[8] = c(4, 2);
      pqc[9] = c(5, 0); pqc[10] = c(5, 1); pqc[11] = c(5, 2);
    }

    r->vel._present = true;
    r->vel._value.vx = (*context)->state.v()(0);
    r->vel._value.vy = (*context)->state.v()(1);
    r->vel._value.vz = (*context)->state.v()(2);
    r->avel._present = true;
    r->avel._value.wx = (*context)->state.w()(0);
    r->avel._value.wy = (*context)->state.w()(1);
    r->avel._value.wz = (*context)->state.w()(2);
    {
      Eigen::Matrix<double, 6, 6> c =
        (*context)->state.vw_cov((*context)->cov);

      double *vc = r->vel_cov._value.cov;
      r->vel_cov._present = true;
      vc[0] = c(0, 0);
      vc[1] = c(1, 0);  vc[2] = c(1, 1);
      vc[3] = c(2, 0);  vc[4] = c(2, 1);  vc[5] = c(2, 2);

      double *wc = r->avel_cov._value.cov;
      r->avel_cov._present = true;
      wc[0] = c(0, 0);
      wc[1] = c(1, 0);  wc[2] = c(1, 1);
      wc[3] = c(2, 0);  wc[4] = c(2, 1);  wc[5] = c(2, 2);
    }

    r->acc._present = true;
    r->acc._value.ax = (*context)->state.a()(0);
    r->acc._value.ay = (*context)->state.a()(1);
    r->acc._value.az = (*context)->state.a()(2);
    r->aacc._present = false;
    {
      Eigen::Matrix<double, 3, 3> c =
        (*context)->state.a_cov((*context)->cov);

      double *ac = r->acc_cov._value.cov;
      r->acc_cov._present = true;
      ac[0] = c(0, 0);
      ac[1] = c(1, 0);  ac[2] = c(1, 1);
      ac[3] = c(2, 0);  ac[4] = c(2, 1);  ac[5] = c(2, 2);

      r->aacc_cov._present = false;
    }
  }
  frame->write("robot", self);

  /* log */
  if ((*log_state)->req.aio_fildes >= 0) {
    (*log_state)->total++;
    if ((*log_state)->total % (*log_state)->decimation == 0) {
      if ((*log_state)->pending) {
        if (aio_error(&(*log_state)->req) != EINPROGRESS) {
          (*log_state)->pending = false;
          if (aio_return(&(*log_state)->req) <= 0) {
            warn("log_state");
            close((*log_state)->req.aio_fildes);
            (*log_state)->req.aio_fildes = -1;
          }
        } else {
          (*log_state)->skipped = true;
          (*log_state)->missed++;
        }
      }
    }

    if ((*log_state)->req.aio_fildes >= 0 && !(*log_state)->pending) {
      const Eigen::Matrix<double, 3, 1> rpy = (*context)->state.rpy();
      const Eigen::Matrix<double, 3, 3> rpy_cov =
        (*context)->state.rpy_cov((*context)->cov);

      (*log_state)->req.aio_nbytes = snprintf(
        (*log_state)->buffer, sizeof((*log_state)->buffer),
        "%s" pom_log_line_fmt "\n",
        (*log_state)->skipped ? "\n" : "",
        r->ts.sec, r->ts.nsec,
        r->intrinsic,
        r->pos._present, r->att._present,
        r->vel._present, r->avel._present,
        r->acc._present,

        r->pos._value.x, r->pos._value.y, r->pos._value.z,
        rpy(0), rpy(1), rpy(2),
        r->vel._value.vx, r->vel._value.vy, r->vel._value.vz,
        r->avel._value.wx, r->avel._value.wy, r->avel._value.wz,
        r->acc._value.ax, r->acc._value.ay, r->acc._value.az,

        r->pos_cov._value.cov[0],
        r->pos_cov._value.cov[2],
        r->pos_cov._value.cov[5],
        rpy_cov(0,0),
        rpy_cov(1,1),
        rpy_cov(2,2),
        r->vel_cov._value.cov[0],
        r->vel_cov._value.cov[2],
        r->vel_cov._value.cov[5],
        r->avel_cov._value.cov[0],
        r->avel_cov._value.cov[2],
        r->avel_cov._value.cov[5],
        r->acc_cov._value.cov[0],
        r->acc_cov._value.cov[2],
        r->acc_cov._value.cov[5]);

      if (aio_write(&(*log_state)->req)) {
        warn("log_state");
        close((*log_state)->req.aio_fildes);
        (*log_state)->req.aio_fildes = -1;
      } else
        (*log_state)->pending = true;

      (*log_state)->skipped = false;
    }
  }

  return pom_pause_exec;
}
